/**
 * Copyright (c) Members of the EMI Collaboration. 2011.
 * See http://eu-emi.eu/partners/ for details on the copyright holders.
 * For license conditions see http://www.apache.org/licenses/LICENSE-2.0
 */
package org.glite.pseudo.common.http;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.GeneralSecurityException;
import java.util.Properties;

import org.apache.http.client.HttpClient;
import org.apache.http.client.protocol.RequestAcceptEncoding;
import org.apache.http.client.protocol.ResponseContentEncoding;
import org.apache.http.conn.ClientConnectionManager;
import org.apache.http.conn.scheme.PlainSocketFactory;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.scheme.SchemeRegistry;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.conn.SingleClientConnManager;
import org.apache.http.impl.conn.tsccm.ThreadSafeClientConnManager;
import org.glite.security.trustmanager.ContextWrapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A builder for {@link HttpClient}s.
 * 
 * This builder will produce clients that employ the {@link MultiThreadedHttpConnectionManager} and as such users of the
 * clients MUST be sure to invoke {@link org.apache.commons.httpclient.HttpMethod#releaseConnection()} after they have
 * finished with the method.
 * 
 * This class is partly based on <code>org.opensaml.ws.soap.client.http.HttpClientBuilder</code>.
 */
public class HttpClientBuilder {

	/** Logging */
    private static final Logger log = LoggerFactory.getLogger(HttpClientBuilder.class);
    
    /** The URL for the server's end-point used for the HttpClient to be built. */
    private URL serverUrl;
    
    /** Determines whether to pool connections to a host. Default value: true */
    private boolean connectionPooling;

    /** Total number of connections that may be open. Default value: 20 */
    private int connectionsMaxTotal;

    /** Maximum number of connections that may be opened to a single host. Default value: 2 */
    private int connectionsMaxPerRoute;
    
    /** Properties of the connection for the HttpClient to be built. */
    private Properties connectionProperties = System.getProperties();
    
    /**
     * Constructs a new <code>HttpClientBuilder</code>.
     * @param url to the server's end-point.
     * @param gridProxyFile the full path for the proxy file.
     */
    public HttpClientBuilder(URL url, String gridProxyFile) {
    	this.serverUrl = url;
    	this.setGridProxyFile(gridProxyFile);
    }
    
    /**
     * Constructs a new <code>HttpClientBuilder</code>.
     * @param url to the server's end-point as <code>String</code>.
     * @param gridProxyFile the full path for the proxy file.
     */
    public HttpClientBuilder(String url, String gridProxyFile) {
    	this.setUrl(url);
    	this.setGridProxyFile(gridProxyFile);
    }
    
    /**
     * Constructs a new <code>HttpClientBuilder</code>.
     * @param url to the server's end-point as <code>String</code>.
     * @param certFile the full path for the client certificate file.
     * @param keyFile the full path for the unencrypted private key file corresponding to the certificate.
     */
    public HttpClientBuilder(String url, String certFile, String keyFile) {
    	this(url, certFile, keyFile, null);
    }
    
    /**
     * Constructs a new <code>HttpClientBuilder</code>.
     * @param url to the server's end-point as <code>String</code.
     * @param certFile the full path for the client certificate file.
     * @param keyFile the full path for the encrypted private key file corresponding to the certificate.
     * @param keyPasswd the password used for decrypting the private key.
     */
	public HttpClientBuilder(String url, String certFile, String keyFile, String keyPasswd) {
		this.setUrl(url);
		this.setConnectionProperty("gridProxyFile", null);
    	this.setConnectionProperty("sslCertFile", certFile);
    	this.setConnectionProperty("sslKey", keyFile);
    	if (keyPasswd != null && !keyPasswd.equals("")) {
    		this.connectionProperties.setProperty("sslKeyPasswd", keyPasswd);
    	}
	}
	
	/**
	 * Sets one property for the connection of the HttpClient.
	 * @param key of the property.
	 * @param value for the property.
	 */
	public void setConnectionProperty(String key, String value) {
		log.debug("Setting connection property {} to {}", key, value);
		if (value == null) {
			this.connectionProperties.remove(key);
		} else {
			this.connectionProperties.setProperty(key, value);
		}
	}
	
	/**
	 * Sets the full path for the proxy file.
	 * @param gridProxyFile the full path for the proxy file.
	 */
	protected void setGridProxyFile(String gridProxyFile) {
	   	if (gridProxyFile == null || gridProxyFile.equals("")) {
    		log.error("gridProxyFile cannot be null!");
    	} else {
    		this.setConnectionProperty("sslCertFile", null);
    		this.setConnectionProperty("sslKey", null);
    		this.setConnectionProperty("sslKeyPasswd", null);
    		this.setConnectionProperty("gridProxyFile", gridProxyFile);
    	}		
	}
	
	/**
	 * Sets the server's end-point url.
	 * @param url the end-point url as <code>String</code>.
	 */
	protected void setUrl(String url) {
		log.debug("Setting service URL for {}", url);
		try {
			this.serverUrl = new URL(url);
		} catch (MalformedURLException e) {
			log.error("The given URL {} is not valid", url);
		}
	}

	/**
	 * Builds a new <code>HttpClient</code> with the current parameters.
	 * @return the <code>HttpClient</code>
	 */
	public HttpClient buildClient() throws IOException, GeneralSecurityException {
        final DefaultHttpClient client = new DefaultHttpClient(buildConnectionManager());
        client.addRequestInterceptor(new RequestAcceptEncoding());
        client.addResponseInterceptor(new ResponseContentEncoding());
		
		return client;
	}

    /**
     * Builds the connection manager used by the HTTP client. If {@link #connectionPooling} is false then the
     * {@link SingleClientConnManager} is used. Otherwise the {@link ThreadSafeClientConnManager} is used with
     * {@link ThreadSafeClientConnManager#setDefaultMaxPerRoute(int)} set to {@link #connectionsMaxPerRoute} and
     * {@link ThreadSafeClientConnManager#setMaxTotalConnections(int)} set to {@link #connectionsMaxTotal}.
     * @return the connection manager used by the HttpClient
     */
    private ClientConnectionManager buildConnectionManager() throws IOException, GeneralSecurityException {
        final SchemeRegistry registry = buildSchemeRegistry();

		if (!connectionPooling) {
            return new SingleClientConnManager(registry);
        } else {
            final ThreadSafeClientConnManager manager = new ThreadSafeClientConnManager(registry);
			manager.setDefaultMaxPerRoute(connectionsMaxPerRoute);
            manager.setMaxTotal(connectionsMaxTotal);
            return manager;
        }
    }

	private SchemeRegistry buildSchemeRegistry() throws IOException, GeneralSecurityException {
        final SchemeRegistry registry = new SchemeRegistry();
        String protocol = serverUrl.getProtocol();
        
        if (protocol.equals("http")) {
        	registry.register(new Scheme("http", serverUrl.getPort(), PlainSocketFactory.getSocketFactory()));
        } else if (protocol.equals("https")){

        	final SSLSocketFactory sslSF;
        	ContextWrapper ctxWrapper;
        	ctxWrapper = new ContextWrapper(this.connectionProperties);
        	sslSF = new SSLSocketFactory(ctxWrapper.getContext());
        	registry.register(new Scheme("https", serverUrl.getPort(), sslSF));
        } else {
        	log.error("Invalid protocol ({}), HTTP client could not be initialized!", protocol);
        }

        return registry;
	}

}
